/*
 * sysdb.c  -  Read or write a system definition using the systems database
 *
 * Copyright (C) 2003-2006 Gero Kuhlmann <gero@gkminix.han.de>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * $Id: sysdb.c,v 1.9 2006/03/07 19:26:19 gkminix Exp $
 */

#include <common.h>
#include <nblib.h>
#include "privlib.h"



/*
 * Error handler for database routines
 */
static void syserr __F((dbname, msg, class),
				const char *dbname AND
				const char *msg AND
				dberrclass class)
{
  char *classmsg = NULL;

  /* Print error message */
  switch (class) {
	case DBERR_CLASS_WARNING:
		classmsg = "Warning";
		break;
	case DBERR_CLASS_ERROR:
		classmsg = "Error";
		break;
	case DBERR_CLASS_FATAL:
		classmsg = "Fatal";
		break;
	case DBERR_CLASS_INFO:
		prnlog(LOGLEVEL_INFO, "Info: [%s] %s\n", dbname, msg);
		/* Fall through */
	default:
		classmsg = NULL;
		break;
  }
  if (classmsg != NULL)
	prnerr("%s: [%s] %s", classmsg, dbname, msg);

  /* Terminate in case of fatal error */
  if (class == DBERR_CLASS_FATAL)
	exit(EXIT_DB);
}



/*
 * Read a system definition out of the systems database
 */
void readsysdb __F((recname, params),
				char *recname AND
				struct paramdef *params)
{
  struct sectdef sects[2];
  DBHDL dbhandle;

  /* Check that a system database name has been specified */
  if (sysdbname == NULL) {
	prnerr("system database name missing");
	exit(EXIT_DB);
  }

  /* Generate a system database handle */
  dbhandle = createdb(sysdbname, &syserr);
  if (dbhandle == DBHDL_NULL) {
	prnerr("invalid system database \"%s\"", sysdbname);
	exit(EXIT_CONFIG);
  }

  /* Open systems database */
  if (opendb(dbhandle, TRUE) == 0 || dberrors > 0) {
	prnerr("unable to open database \"%s\"", sysdbname);
	exit(EXIT_DB);
  }

  /* Create a section definition */
  memzero(sects, sizeof(sects));
  sects[0].name = recname;
  sects[0].params = params;
  sects[0].startsect = NULL;
  sects[0].endsect = NULL;

  /* Search for record name */
  if (!findfirst(dbhandle, recname)) {
	if (dberrors == 0)
		prnerr("unable to find record \"%s\" in systems database",
								recname);
	exit(EXIT_DB);
  }

  /* Read all records for this system */
  do {
	if (readrec(dbhandle, sects) != 1) {
		if (dberrors == 0)
			prnerr("error reading record \"%s\" from systems database",
								recname);
		exit(EXIT_DB);
	}
  } while (findnext(dbhandle));
  if (dberrors > 0)
	exit(EXIT_DB);

  /* Close and free everything */
  closedb(dbhandle);
  freedb(dbhandle);
}



/*
 * Write a system definition into the systems database
 *
 * Writing is a bit more complicated than reading. This is because
 * a database definition may contain multiple databases. We search
 * for the last database which is available for writing and which
 * has the "add" flag.
 */
void writesysdb __F((recname, params),
				char *recname AND
				struct paramdef *params)
{
  struct sectdef sects[2];
  DBHDL dbhandle;
  int dbstate;
  char *cp;

  /* Check that the record name is correct */
  if ((cp = strchr(recname, ':')) == NULL ||
      cp[1] == '\0' || cp[1] == '*' || cp[2] == '\0') {
	prnerr("invalid record name \"%s\" for system database", recname);
	exit(EXIT_DB);
  }

  /* Check that a system database name has been specified */
  if (sysdbname == NULL) {
	prnerr("system database name missing");
	exit(EXIT_DB);
  }

  /* Generate a system database handle */
  dbhandle = createdb(sysdbname, &syserr);
  if (dbhandle == DBHDL_NULL) {
	prnerr("invalid system database \"%s\"", sysdbname);
	exit(EXIT_CONFIG);
  }

  /* Open systems database */
  if (opendb(dbhandle, FALSE) == 0 || dberrors > 0) {
	prnerr("unable to open database \"%s\"", sysdbname);
	exit(EXIT_DB);
  }

  /* Create a section definition */
  memzero(sects, sizeof(sects));
  sects[0].name = recname;
  sects[0].params = params;
  sects[0].startsect = NULL;
  sects[0].endsect = NULL;

  /*
   * Delete all records with this name. If we can't delete a record
   * because it's in a read-only database that's not a problem. This
   * delete loop just avoids that a database can grow endlessly.
   */
  if (findfirst(dbhandle, recname)) {
	do {
		dbstate = getstatedb(dbhandle, NULL) &
				(DBSTATUS_RO | DBSTATUS_VALID | DBSTATUS_POS);
		if (dbstate == (DBSTATUS_VALID | DBSTATUS_POS))
			delrec(dbhandle);
		if (dberrors > 0)
			exit(EXIT_DB);
	} while (findnext(dbhandle));
  }
  if (dberrors > 0)
	exit(EXIT_DB);

  /* Now find a database to which we can safely add the new record */
  dbstate = 0;
  if (lastrec(dbhandle)) {
	do {
		dbstate = getstatedb(dbhandle, NULL) &
				(DBSTATUS_RO | DBSTATUS_VALID | DBSTATUS_ADD);
		if (dbstate == (DBSTATUS_VALID | DBSTATUS_ADD))
			break;
	} while (prevdb(dbhandle));
  } else
	dbstate = getstatedb(dbhandle, NULL) &
				(DBSTATUS_RO | DBSTATUS_VALID | DBSTATUS_ADD);
  if (dberrors > 0)
	exit(EXIT_DB);

  /* Check that we can add a new record to the current database */
  if (dbstate != (DBSTATUS_VALID | DBSTATUS_ADD)) {
	prnerr("unable to find a systems database which allows adding");
	exit(EXIT_DB);
  }

  /* Now finally add the new record at the current position */
  addrec(dbhandle, sects, 0);
  if (dberrors > 0)
	exit(EXIT_DB);

  /* Close and free everything */
  closedb(dbhandle);
  freedb(dbhandle);
}

